"""Pin definitions for the Coral Edge TPU Dev board."""

from adafruit_blinka.microcontroller.nxp_imx8m import pin

SDA = pin.I2C2_SDA
SCL = pin.I2C2_SCL

PWM1 = pin.PWM1
PWM2 = pin.PWM2
PWM3 = pin.PWM3

GPIO_P13 = pin.GPIO6
GPIO_P16 = pin.GPIO73
GPIO_P18 = pin.GPIO138
GPIO_P29 = pin.GPIO7
GPIO_P31 = pin.GPIO8
GPIO_P36 = pin.GPIO141
GPIO_P37 = pin.GPIO77

MISO = pin.ECSPI1_MISO
MOSI = pin.ECSPI1_MOSI
SCLK = pin.ECSPI1_SCLK
SCK = pin.ECSPI1_SCLK
SS0 = pin.ECSPI1_SS0




"""
`busio` - Bus protocol support like I2C and SPI
=================================================

See `CircuitPython:busio` in CircuitPython for more details.

* Author(s): cefn
"""

import threading

from adafruit_blinka import Enum, Lockable, agnostic
from adafruit_blinka.agnostic import board_id, detector
import adafruit_platformdetect.board as ap_board

class I2C(Lockable):
    def __init__(self, scl, sda, frequency=400000):
        self.init(scl, sda, frequency)

    def init(self, scl, sda, frequency):
        self.deinit()
        if detector.board.ftdi_ft232h:
            from adafruit_blinka.microcontroller.ft232h.i2c import I2C
            self._i2c = I2C(frequency=frequency)
            return
        elif detector.board.binho_nova:
            from adafruit_blinka.microcontroller.nova.i2c import I2C
        elif detector.board.microchip_mcp2221:
            from adafruit_blinka.microcontroller.mcp2221.i2c import I2C
            self._i2c = I2C(frequency=frequency)
            return
        elif detector.board.any_embedded_linux:
            from adafruit_blinka.microcontroller.generic_linux.i2c import I2C as _I2C
        else:
            from machine import I2C as _I2C
        from microcontroller.pin import i2cPorts
        for portId, portScl, portSda in i2cPorts:
            try:
                if scl == portScl and sda == portSda:
                    self._i2c = _I2C(portId, mode=_I2C.MASTER, baudrate=frequency)
                    break
            except RuntimeError:
                pass
        else:
            raise ValueError(
                "No Hardware I2C on (scl,sda)={}\nValid I2C ports: {}".format((scl, sda), i2cPorts)
            )

        self._lock = threading.RLock()

    def deinit(self):
        try:
            del self._i2c
        except AttributeError:
            pass

    def __enter__(self):
        self._lock.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._lock.release()
        self.deinit()

    def scan(self):
        return self._i2c.scan()

    def readfrom_into(self, address, buffer, *, start=0, end=None):
        if start is not 0 or end is not None:
            if end is None:
                end = len(buffer)
            buffer = memoryview(buffer)[start:end]
        stop = True  # remove for efficiency later
        return self._i2c.readfrom_into(address, buffer, stop=stop)

    def writeto(self, address, buffer, *, start=0, end=None, stop=True):
        if isinstance(buffer, str):
            buffer = bytes([ord(x) for x in buffer])
        if start is not 0 or end is not None:
            if end is None:
                return self._i2c.writeto(address, memoryview(buffer)[start:], stop=stop)
            else:
                return self._i2c.writeto(address, memoryview(buffer)[start:end], stop=stop)
        return self._i2c.writeto(address, buffer, stop=stop)

    def writeto_then_readfrom(self, address, buffer_out, buffer_in, *, out_start=0, out_end=None, in_start=0, in_end=None, stop=False):
        return self._i2c.writeto_then_readfrom(address, buffer_out, buffer_in,
                                               out_start=out_start, out_end=out_end,
                                               in_start=in_start, in_end=in_end, stop=stop)

class SPI(Lockable):
    def __init__(self, clock, MOSI=None, MISO=None):
        self.deinit()
        if detector.board.ftdi_ft232h:
            from adafruit_blinka.microcontroller.ft232h.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.ft232h.pin import SCK, MOSI, MISO
            self._spi = _SPI()
            self._pins = (SCK, MOSI, MISO)
            return
        elif detector.board.binho_nova:
            from adafruit_blinka.microcontroller.nova.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.nova.pin import SCK, MOSI, MISO
            self._spi = _SPI(clock)
            self._pins = (SCK, MOSI, MISO)
            return
        elif detector.board.any_embedded_linux:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        else:
            from machine import SPI as _SPI
        from microcontroller.pin import spiPorts
        for portId, portSck, portMosi, portMiso in spiPorts:
            if ((clock == portSck) and                   # Clock is required!
                (MOSI == portMosi or MOSI == None) and   # But can do with just output
                (MISO == portMiso or MISO == None)):      # Or just input
                self._spi = _SPI(portId)
                self._pins = (portSck, portMosi, portMiso)
                break
        else:
            raise ValueError(
                "No Hardware SPI on (SCLK, MOSI, MISO)={}\nValid SPI ports:{}".
                format((clock, MOSI, MISO), spiPorts))

    def configure(self, baudrate=100000, polarity=0, phase=0, bits=8):
        if detector.board.any_raspberry_pi or detector.board.any_raspberry_pi_40_pin:
            from adafruit_blinka.microcontroller.bcm283x.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif detector.board.any_beaglebone:
            from adafruit_blinka.microcontroller.am335x.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.ORANGE_PI_PC or board_id == ap_board.ORANGE_PI_R1 or board_id == ap_board.ORANGE_PI_ZERO:
            from adafruit_blinka.microcontroller.allwinner_h3.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.GIANT_BOARD:
            from adafruit_blinka.microcontroller.sama5.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.CORAL_EDGE_TPU_DEV:
            from adafruit_blinka.microcontroller.nxp_imx8m.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.ODROID_C2:
            from adafruit_blinka.microcontroller.amlogic.s905.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.DRAGONBOARD_410C:
            from adafruit_blinka.microcontroller.snapdragon.apq8016.pin import Pin
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
        elif board_id == ap_board.JETSON_NANO:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.tegra.t210.pin import Pin
        elif board_id == ap_board.JETSON_TX1:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.tegra.t210.pin import Pin
        elif board_id == ap_board.JETSON_TX2:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.tegra.t186.pin import Pin
        elif board_id == ap_board.JETSON_XAVIER:
            from adafruit_blinka.microcontroller.generic_linux.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.tegra.t194.pin import Pin
        elif detector.board.ftdi_ft232h:
            from adafruit_blinka.microcontroller.ft232h.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.ft232h.pin import Pin
        elif detector.board.binho_nova:
            from adafruit_blinka.microcontroller.nova.spi import SPI as _SPI
            from adafruit_blinka.microcontroller.nova.pin import Pin
        else:
            from machine import SPI as _SPI
            from machine import Pin

        if self._locked:
            # TODO check if #init ignores MOSI=None rather than unsetting, to save _pinIds attribute
            self._spi.init(
                baudrate=baudrate,
                polarity=polarity,
                phase=phase,
                bits=bits,
                firstbit=_SPI.MSB,
                sck=Pin(self._pins[0].id),
                mosi=Pin(self._pins[1].id),
                miso=Pin(self._pins[2].id)
            )
        else:
            raise RuntimeError("First call try_lock()")

    def deinit(self):
        self._spi = None
        self._pinIds = None

    @property
    def frequency(self):
        try:
            return self._spi.frequency
        except AttributeError:
            raise NotImplementedError("Frequency attribute not implemented for this platform")

    def write(self, buf, start=0, end=None):
        return self._spi.write(buf, start, end)

    def readinto(self, buf, start=0, end=None, write_value=0):
        return self._spi.readinto(buf, start, end, write_value=write_value)

    def write_readinto(self, buffer_out, buffer_in,  out_start=0, out_end=None, in_start=0, in_end=None):
        return self._spi.write_readinto(buffer_out, buffer_in, out_start, out_end, in_start, in_end)


class UART(Lockable):
    class Parity(Enum):
        pass

    Parity.ODD = Parity()
    Parity.EVEN = Parity()

    def __init__(self,
                 tx,
                 rx,
                 baudrate=9600,
                 bits=8,
                 parity=None,
                 stop=1,
                 timeout=1000,
                 receiver_buffer_size=64,
                 flow=None):
        if detector.board.any_embedded_linux:
            raise RuntimeError('busio.UART not supported on this platform. Please use pyserial instead.')
        elif detector.board.binho_nova:
            from adafruit_blinka.microcontroller.nova.uart import UART as _UART
        else:
            from machine import UART as _UART

        if detector.board.binho_nova:
            from adafruit_blinka.microcontroller.nova.pin import uartPorts
        else:
            from microcontroller.pin import uartPorts

        self.baudrate = baudrate

        if flow is not None:  # default 0
            raise NotImplementedError(
                "Parameter '{}' unsupported on {}".format(
                    "flow", agnostic.board_id))

        # translate parity flag for Micropython
        if parity is UART.Parity.ODD:
            parity = 1
        elif parity is UART.Parity.EVEN:
            parity = 0
        elif parity is None:
            pass
        else:
            raise ValueError("Invalid parity")

        # check tx and rx have hardware support
        for portId, portTx, portRx in uartPorts:  #
            if portTx == tx and portRx == rx:
                self._uart = _UART(
                    portId,
                    baudrate,
                    bits=bits,
                    parity=parity,
                    stop=stop,
                    timeout=timeout,
                    read_buf_len=receiver_buffer_size
                )
                break
        else:
            raise ValueError(
                "No Hardware UART on (tx,rx)={}\nValid UART ports: {}".format((tx, rx), uartPorts)
            )

    def deinit(self):
        if detector.board.binho_nova:
            self._uart.deinit()
        self._uart = None

    def read(self, nbytes=None):
        return self._uart.read(nbytes)

    def readinto(self, buf, nbytes=None):
        return self._uart.readinto(buf, nbytes)

    def readline(self):
        return self._uart.readline()

    def write(self, buf):
        return self._uart.write(buf)



def const(x):
    return x


def native(f):
    return f


def viper(f):
    raise SyntaxError("invalid micropython decorator")


def asm_thumb(f):
    raise SyntaxError("invalid micropython decorator")


import time
import hid

# from the C driver
# http://ww1.microchip.com/downloads/en/DeviceDoc/mcp2221_0_1.tar.gz
# others (???) determined during driver developement
# pylint: disable=bad-whitespace
RESP_ERR_NOERR              = 0x00
RESP_ADDR_NACK              = 0x25
RESP_READ_ERR               = 0x7F
RESP_READ_COMPL             = 0x55
RESP_READ_PARTIAL           = 0x54 # ???
RESP_I2C_IDLE               = 0x00
RESP_I2C_START_TOUT         = 0x12
RESP_I2C_RSTART_TOUT        = 0x17
RESP_I2C_WRADDRL_TOUT       = 0x23
RESP_I2C_WRADDRL_WSEND      = 0x21
RESP_I2C_WRADDRL_NACK       = 0x25
RESP_I2C_WRDATA_TOUT        = 0x44
RESP_I2C_RDDATA_TOUT        = 0x52
RESP_I2C_STOP_TOUT          = 0x62

RESP_I2C_MOREDATA           = 0x43 # ???
RESP_I2C_PARTIALDATA        = 0x41 # ???
RESP_I2C_WRITINGNOSTOP      = 0x45 # ???

MCP2221_RETRY_MAX           = 50
MCP2221_MAX_I2C_DATA_LEN    = 60
MASK_ADDR_NACK              = 0x40
# pylint: enable=bad-whitespace

class MCP2221:

    VID = 0x04D8
    PID = 0x00DD

    GP_GPIO = 0b000
    GP_DEDICATED = 0b001
    GP_ALT0 = 0b010
    GP_ALT1 = 0b011
    GP_ALT2 = 0b100

    def __init__(self):
        self._hid = hid.device()
        self._hid.open(MCP2221.VID, MCP2221.PID)
        self._reset()
        time.sleep(0.25)

    def _hid_xfer(self, report, response=True):
        # first byte is report ID, which =0 for MCP2221
        # remaing bytes = 64 byte report data
        # https://github.com/libusb/hidapi/blob/083223e77952e1ef57e6b77796536a3359c1b2a3/hidapi/hidapi.h#L185
        self._hid.write(b'\0' + report + b'\0'*(64-len(report)))
        if response:
            # return is 64 byte response report
            return self._hid.read(64)

    #----------------------------------------------------------------
    # MISC
    #----------------------------------------------------------------
    def gp_get_mode(self, pin):
        return self._hid_xfer(b'\x61')[22+pin] & 0x07

    def gp_set_mode(self, pin, mode):
        # get current settings
        current = self._hid_xfer(b'\x61')
        # empty report, this is safe since 0's = no change
        report = bytearray(b'\x60'+b'\x00'*63)
        # set the alter GP flag byte
        report[7] = 0xFF
        # each pin can be set individually
        # but all 4 get set at once, so we need to
        # transpose current settings
        report[8]  = current[22]  # GP0
        report[9]  = current[23]  # GP1
        report[10] = current[24]  # GP2
        report[11] = current[25]  # GP3
        # then change only the one
        report[8+pin] = mode & 0x07
        # and make it so
        self._hid_xfer(report)

    def _pretty_report(self, report):
        print("     0  1  2  3  4  5  6  7  8  9")
        index = 0
        for row in range(7):
            print("{} : ".format(row), end='')
            for _ in range(10):
                print("{:02x} ".format(report[index]), end='')
                index += 1
                if index > 63:
                    break
            print()

    def _status_dump(self):
        self._pretty_report(self._hid_xfer(b'\x10'))

    def _sram_dump(self):
        self._pretty_report(self._hid_xfer(b'\x61'))

    def _reset(self):
        self._hid_xfer(b'\x70\xAB\xCD\xEF', response=False)
        start = time.monotonic()
        while time.monotonic() - start < 5:
            try:
                self._hid.open(MCP2221.VID, MCP2221.PID)
            except OSError:
                # try again
                time.sleep(0.1)
                continue
            return
        raise OSError("open failed")

    #----------------------------------------------------------------
    # GPIO
    #----------------------------------------------------------------
    def gpio_set_direction(self, pin, mode):
        report = bytearray(b'\x50'+b'\x00'*63)  # empty set GPIO report
        offset = 4 * (pin + 1)
        report[offset] = 0x01                   # set pin direction
        report[offset+1] = mode                 # to this
        self._hid_xfer(report)

    def gpio_set_pin(self, pin, value):
        report = bytearray(b'\x50'+b'\x00'*63)  # empty set GPIO report
        offset = 2 + 4 * pin
        report[offset] = 0x01                   # set pin value
        report[offset+1] = value                # to this
        self._hid_xfer(report)

    def gpio_get_pin(self, pin):
        resp = self._hid_xfer(b'\x51')
        offset = 2 + 2 * pin
        if resp[offset] == 0xEE:
            raise RuntimeError("Pin is not set for GPIO operation.")
        else:
            return resp[offset]

    #----------------------------------------------------------------
    # I2C
    #----------------------------------------------------------------
    def _i2c_status(self):
        resp = self._hid_xfer(b'\x10')
        if resp[1] != 0:
            raise RuntimeError("Couldn't get I2C status")
        return resp

    def _i2c_state(self):
        return self._i2c_status()[8]

    def _i2c_cancel(self):
        resp = self._hid_xfer(b'\x10\x00\x10')
        if resp[1] != 0x00:
            raise RuntimeError("Couldn't cancel I2C")
        if resp[2] == 0x10:
            # bus release will need "a few hundred microseconds"
            time.sleep(0.001)

    def _i2c_write(self, cmd, address, buffer, start=0, end=None):
        if self._i2c_state() != 0x00:
            self._i2c_cancel()

        end = end if end else len(buffer)
        length = end - start
        retries = 0

        while (end - start) > 0:
            chunk = min(end - start, MCP2221_MAX_I2C_DATA_LEN)
            # write out current chunk
            resp = self._hid_xfer(bytes([cmd,
                                         length & 0xFF,
                                         (length >> 8) & 0xFF,
                                         address << 1]) +
                                         buffer[start:(start+chunk)])
            # check for success
            if resp[1] != 0x00:
                if resp[2] in (RESP_I2C_START_TOUT,
                               RESP_I2C_WRADDRL_TOUT,
                               RESP_I2C_WRADDRL_NACK,
                               RESP_I2C_WRDATA_TOUT,
                               RESP_I2C_STOP_TOUT):
                    raise RuntimeError("Unrecoverable I2C state failure")
                retries += 1
                if retries >= MCP2221_RETRY_MAX:
                    raise RuntimeError("I2C write error, max retries reached.")
                time.sleep(0.001)
                continue # try again
            # yay chunk sent!
            while self._i2c_state() == RESP_I2C_PARTIALDATA:
                time.sleep(0.001)
            start += chunk
            retries = 0

        # check status in another loop
        for _ in range(MCP2221_RETRY_MAX):
            status = self._i2c_status()
            if status[20] & MASK_ADDR_NACK:
                raise RuntimeError("I2C slave address was NACK'd")
            usb_cmd_status = status[8]
            if usb_cmd_status == 0:
                break
            if usb_cmd_status == RESP_I2C_WRITINGNOSTOP and cmd == 0x94:
                break   # this is OK too!
            if usb_cmd_status in (RESP_I2C_START_TOUT,
                                  RESP_I2C_WRADDRL_TOUT,
                                  RESP_I2C_WRADDRL_NACK,
                                  RESP_I2C_WRDATA_TOUT,
                                  RESP_I2C_STOP_TOUT):
                raise RuntimeError("Unrecoverable I2C state failure")
            time.sleep(0.001)
        else:
            raise RuntimeError("I2C write error: max retries reached.")
        # whew success!

    def _i2c_read(self, cmd, address, buffer, start=0, end=None):
        if self._i2c_state() not in (RESP_I2C_WRITINGNOSTOP, 0):
            self._i2c_cancel()

        end = end if end else len(buffer)
        length = end - start

        # tell it we want to read
        resp = self._hid_xfer(bytes([cmd,
                                     length & 0xFF,
                                     (length >> 8) & 0xFF,
                                     (address << 1) | 0x01]))

        # check for success
        if resp[1] != 0x00:
            raise RuntimeError("Unrecoverable I2C read failure")

        # and now the read part
        while (end - start) > 0:
            for retry in range(MCP2221_RETRY_MAX):
                # the actual read
                resp = self._hid_xfer(b'\x40')
                # check for success
                if resp[1] == RESP_I2C_PARTIALDATA:
                    time.sleep(0.001)
                    continue
                if resp[1] != 0x00:
                    raise RuntimeError("Unrecoverable I2C read failure")
                if resp[2] == RESP_ADDR_NACK:
                    raise RuntimeError("I2C NACK")
                if resp[3] == 0x00 and resp[2] == 0x00:
                    break
                if resp[3] == RESP_READ_ERR:
                    time.sleep(0.001)
                    continue
                if resp[2] in (RESP_READ_COMPL, RESP_READ_PARTIAL):
                    break

            # move data into buffer
            chunk = min(end - start, 60)
            for i, k in enumerate(range(start, start+chunk)):
                buffer[k] = resp[4 + i]
            start += chunk

    def i2c_configure(self, baudrate=100000):
        self._hid_xfer(bytes([0x10,  # set parameters
                              0x00,  # don't care
                              0x00,  # no effect
                              0x20,  # next byte is clock divider
                              12000000 // baudrate - 3]))

    def i2c_writeto(self, address, buffer, *, start=0, end=None):
        self._i2c_write(0x90, address, buffer, start, end)

    def i2c_readfrom_into(self, address, buffer, *, start=0, end=None):
        self._i2c_read(0x91, address, buffer, start, end)

    def i2c_writeto_then_readfrom(self, address, out_buffer, in_buffer, *,
                                  out_start=0, out_end=None,
                                  in_start=0, in_end=None):
        self._i2c_write(0x94, address, out_buffer, out_start, out_end)
        self._i2c_read(0x93, address, in_buffer, in_start, in_end)

    def i2c_scan(self, *, start=0, end=0x79):
        found = []
        for addr in range(start, end+1):
            # try a write
            try:
                self.i2c_writeto(addr, b'\x00')
            except RuntimeError: # no reply!
                continue
            # store if success
            found.append(addr)
        return found

    #----------------------------------------------------------------
    # ADC
    #----------------------------------------------------------------
    def adc_configure(self, vref=0):
        report = bytearray(b'\x60'+b'\x00'*63)
        report[5] = 1 << 7 | (vref & 0b111)
        self._hid_xfer(report)

    def adc_read(self, pin):
        resp = self._hid_xfer(b'\x10')
        return resp[49 + 2 * pin] << 8 | resp[48 + 2 * pin]

    #----------------------------------------------------------------
    # DAC
    #----------------------------------------------------------------
    def dac_configure(self, vref=0):
        report = bytearray(b'\x60'+b'\x00'*63)
        report[3] = 1 << 7 | (vref & 0b111)
        self._hid_xfer(report)

    def dac_write(self, pin, value):
        report = bytearray(b'\x60'+b'\x00'*63)
        report[4] = 1 << 7 | (value & 0b11111)
        self._hid_xfer(report)

mcp2221 = MCP2221()
